use byteorder::{BigEndian, ByteOrder, LittleEndian};
use std::collections::VecDeque;
use std::time::{SystemTime, UNIX_EPOCH};

const FILTER_SIZE: usize = 5;

#[allow(dead_code)]
#[derive(Debug, PartialEq)]
pub enum FaceBlendShape {
    EyeBlinkLeft = 0,
    EyeLookDownLeft = 1,
    EyeLookInLeft = 2,
    EyeLookOutLeft = 3,
    EyeLookUpLeft = 4,
    EyeSquintLeft = 5,
    EyeWideLeft = 6,
    EyeBlinkRight = 7,
    EyeLookDownRight = 8,
    EyeLookInRight = 9,
    EyeLookOutRight = 10,
    EyeLookUpRight = 11,
    EyeSquintRight = 12,
    EyeWideRight = 13,
    JawForward = 14,
    JawLeft = 15,
    JawRight = 16,
    JawOpen = 17,
    MouthClose = 18,
    MouthFunnel = 19,
    MouthPucker = 20,
    MouthLeft = 21,
    MouthRight = 22,
    MouthSmileLeft = 23,
    MouthSmileRight = 24,
    MouthFrownLeft = 25,
    MouthFrownRight = 26,
    MouthDimpleLeft = 27,
    MouthDimpleRight = 28,
    MouthStretchLeft = 29,
    MouthStretchRight = 30,
    MouthRollLower = 31,
    MouthRollUpper = 32,
    MouthShrugLower = 33,
    MouthShrugUpper = 34,
    MouthPressLeft = 35,
    MouthPressRight = 36,
    MouthLowerDownLeft = 37,
    MouthLowerDownRight = 38,
    MouthUpperUpLeft = 39,
    MouthUpperUpRight = 40,
    BrowDownLeft = 41,
    BrowDownRight = 42,
    BrowInnerUp = 43,
    BrowOuterUpLeft = 44,
    BrowOuterUpRight = 45,
    CheekPuff = 46,
    CheekSquintLeft = 47,
    CheekSquintRight = 48,
    NoseSneerLeft = 49,
    NoseSneerRight = 50,
    TongueOut = 51,
    HeadYaw = 52,
    HeadPitch = 53,
    HeadRoll = 54,
    LeftEyeYaw = 55,
    LeftEyePitch = 56,
    LeftEyeRoll = 57,
    RightEyeYaw = 58,
    RightEyePitch = 59,
    RightEyeRoll = 60,
}

#[derive(Debug)]
pub struct LiveLinkFace {
    pub uuid: String,
    pub name: String,
    pub fps: i32,
    version: u32,
    frames: u32,
    sub_frame: i32,
    denominator: i32,
    blend_shapes: Vec<f32>,
    old_blend_shapes: Vec<VecDeque<f32>>,
}

impl LiveLinkFace {
    pub fn new(name: String, uuid: String, fps: i32) -> Self {
        let now = SystemTime::now()
            .duration_since(UNIX_EPOCH)
            .expect("Time went backwards");
        let frames = now.as_secs() as u32;
        let sub_frame = 1056060032; // I don't know why
        let denominator = fps / 60; // 1 most of the time

        let mut old_blend_shapes: Vec<VecDeque<f32>> = Vec::with_capacity(61);
        for _ in 0..61 {
            old_blend_shapes.push(VecDeque::with_capacity(FILTER_SIZE));
        }

        LiveLinkFace {
            uuid,
            name,
            fps,
            version: 6,
            frames,
            sub_frame,
            denominator,
            blend_shapes: vec![0.0; 61],
            old_blend_shapes,
        }
    }

    pub fn encode(&self) -> Vec<u8> {
        let mut encoded: Vec<u8> = Vec::new();
        encoded.extend(&(self.version).to_le_bytes());
        encoded.extend(self.uuid.as_bytes());
        encoded.extend(&(self.name.len() as i32).to_be_bytes());
        encoded.extend(self.name.as_bytes());

        // Prepare frame data
        let now = SystemTime::now()
            .duration_since(UNIX_EPOCH)
            .expect("Time went backwards");
        let frame_time = now.as_secs() as u32;
        let frame_rate = (self.fps, self.denominator);
        let blend_shapes_data = self.blend_shapes.to_vec();

        encoded.extend(&(frame_time).to_be_bytes());
        encoded.extend(&(self.sub_frame).to_be_bytes());
        encoded.extend(&(frame_rate).0.to_be_bytes());
        encoded.extend(&(frame_rate).1.to_be_bytes());
        encoded.extend(&(blend_shapes_data.len() as i32).to_be_bytes());
        for f in &blend_shapes_data {
            encoded.extend(&f.to_be_bytes());
        }

        encoded
    }

    pub fn get_blendshape(&self, index: FaceBlendShape) -> f32 {
        self.blend_shapes[index as usize]
    }

    pub fn set_blendshape(&mut self, index: FaceBlendShape, value: f32, no_filter: bool) {
        let idx = index as usize;
        if no_filter {
            self.blend_shapes[idx] = value;
        } else {
            self.old_blend_shapes[idx].push_back(value);
            let filtered_value: f32 = self.old_blend_shapes[idx].iter().sum::<f32>()
                / self.old_blend_shapes[idx].len() as f32;
            self.blend_shapes[idx] = filtered_value;
        }
    }

    pub fn decode(data: &[u8]) -> (bool, LiveLinkFace) {
        let version = LittleEndian::read_i16(&data[0..4]);
        let uuid = String::from_utf8_lossy(&data[4..41]).to_string();
        let name_length = BigEndian::read_i32(&data[41..45]) as usize;
        let name_end_pos = 45 + name_length;
        let name = String::from_utf8_lossy(&data[45..name_end_pos]).to_string();

        if data.len() > name_end_pos + 16 {
            let frame_number = BigEndian::read_i32(&data[name_end_pos..name_end_pos + 4]) as usize;
            let sub_frame = BigEndian::read_i32(&data[name_end_pos + 4..name_end_pos + 8]);
            let fps = BigEndian::read_i32(&data[name_end_pos + 8..name_end_pos + 12]);
            let denominator = BigEndian::read_i32(&data[name_end_pos + 12..name_end_pos + 16]);
            let (int_bytes, _rest) = data[70..74].split_at(std::mem::size_of::<i8>());
            let data_length = i8::from_be_bytes(int_bytes.try_into().unwrap());

            if data_length != 61 {
                log::debug!(
                    "Blendshape data is malformed. got: {}; requred: 61",
                    data_length
                );
            }

            let mut blend_shapes: [f32; 61] = [0.0; 61];
            for i in 0..61 {
                blend_shapes[i] = BigEndian::read_f32(
                    &data[name_end_pos + 17 + i * 4..name_end_pos + 17 + (i + 1) * 4],
                );
            }

            let mut live_link_face = LiveLinkFace::new(name, uuid, fps);
            live_link_face.version = version as u32;
            live_link_face.frames = frame_number as u32;
            live_link_face.sub_frame = sub_frame;
            live_link_face.denominator = denominator;
            live_link_face.blend_shapes = blend_shapes.to_vec();

            return (true, live_link_face);
        }

        (
            false,
            LiveLinkFace::new("Default".to_string(), "Default".to_string(), 60),
        )
    }
}

#[cfg(test)]
mod tests {
    use crate::{LiveLinkFace, FaceBlendShape};

    const BUF: [u8;315] = [
            0x06, 0x00, 0x00, 0x00, 0x24, 0x30, 0x30, 0x46, 0x44, 0x45, 0x31, 0x43, 0x33, 0x2D,
            0x46, 0x41, 0x38, 0x35, 0x2D, 0x34, 0x30, 0x36, 0x37, 0x2D, 0x39, 0x37, 0x44, 0x38,
            0x2D, 0x31, 0x31, 0x33, 0x37, 0x43, 0x41, 0x31, 0x38, 0x33, 0x30, 0x35, 0x45, 0x00,
            0x00, 0x00, 0x09, 0x46, 0x61, 0x63, 0x65, 0x54, 0x72, 0x61, 0x63, 0x6B, 0x00, 0x00,
            0xFD, 0x2F, 0x3E, 0x02, 0xF6, 0x00, 0x00, 0x00, 0x00, 0x3C, 0x00, 0x00, 0x00, 0x01,
            0x3D, 0x39, 0xC0, 0x37, 0x00, 0x3D, 0x95, 0x3A, 0xAE, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x39, 0xC0, 0x95, 0xF3, 0x3D, 0x95, 0x2D, 0x4E, 0x3D, 0x02, 0xA5, 0x4E, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x39, 0x7B, 0x26, 0xB0, 0x3B, 0x0D, 0x5D, 0x4F, 0x00, 0x00, 0x00, 0x00, 0x3B,
            0x1D, 0xB8, 0x29, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x3B, 0x3B, 0xE3, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x3A, 0xFD, 0xE5,
            0x7A, 0x3B, 0x09, 0x19, 0x78, 0x3B, 0x64, 0x88, 0xAE, 0x3B, 0x2E, 0x7D, 0xE2, 0x00,
            0x00, 0x00, 0x00, 0x3A, 0xD8, 0xD2, 0xAF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xBD,
            0xE8, 0xCC, 0x08, 0xBE, 0x39, 0x6A, 0x4C, 0x3C, 0x36, 0xE9, 0x45, 0xBC, 0x9D, 0xC5,
            0x24, 0x3D, 0x2A, 0x20, 0x39, 0xBA, 0xE9, 0xF6, 0xC7, 0xBC, 0x9B, 0x6C, 0xD2, 0x3D,
            0x2A, 0x1E, 0x4B, 0x3A, 0xA1, 0xB0, 0x42,
        ];
    #[test]
    fn decode_buf() {
        let face_data = LiveLinkFace::decode(&BUF);

        assert!(face_data.0);
        assert_eq!(face_data.1.get_blendshape(FaceBlendShape::EyeBlinkLeft), 0.00036662072);
    }
}

